/*
 * ministream - minimal (subset) appstream metadata parser
 * Copyright (c) 2026  Florian Leander Singer <sp1rit@disroot.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __MINISTREAM_PRIV_H__
#define __MINISTREAM_PRIV_H__

#include "ministream.h"

G_BEGIN_DECLS

struct _MsRelease {
	gchar* version;
	gchar* description;
};

void ms_release_free(MsRelease* self);

struct _MsComponent {
	gatomicrefcount rc;

	gchar* id;
	gchar* name;
	gchar* homepage_uri;
	gchar* bugtracker_uri;
	gchar* help_uri;
	GPtrArray* desktop_launchables;
	GPtrArray* releases;
	gchar* project_license;
	gchar* developer_name;
};

gchar* ms_get_current_locale(void);
guint ms_locale_score(const gchar* lhs, const gchar* rhs);

gint ms_vercmp(const gchar* lhs, const gchar* rhs);

G_END_DECLS

#endif // __MINISTREAM_PRIV_H__
